#ifndef RZB_ALERT_DB_H
#define RZB_ALERT_DB_H

#include <uuid/uuid.h>
#include "rzb_global.h"
#include "rzb_alert_global.h"

/* Definitions to Support ALERT_TRACKER */
#define FULL_BLOCK_STRING  "Full_Block"
#define NORM_BLOCK_STRING  "Norm_Block"
#define SHORT_DATA_STRING  "Short_Data"
#define LONG_DATA_STRING   "Long_Data"
#define FULL_ALERT_FIELD   "Data_ID"
#define NORM_ALERT_FIELD   "Norm_ID"

typedef struct _ALERT_TRACKER
{
    unsigned char *main_md5;            // Describe the original data block
    unsigned      main_size;
    const char    *blob_string;         // "Full_Block", "Norm_Block", "Short_Data"
    unsigned      blob_string_len;
    unsigned      blob_type;            // FULL_BLOCK, NORM_BLOCK, SHORT_DATA
    const char    *blob_alert_field;    // "Data_ID", "Norm_ID" (fields in alert table);
    unsigned      blob_alert_field_len;
    uuid_t        data_type;
    unsigned char *blob_md5;            // Describe the data block currently being evaluated
    unsigned      blob_size;
    void          *blob_data;
    unsigned      block_id;
    unsigned      data_component_id;
    unsigned      current_component_id;
    unsigned      alert_id;             // Track the alert id we're commiting against
    RZBNetworkConnection *net;
} ALERT_TRACKER;

// Function Declarations for General Alerting
HRESULT insertAlert(ALERT_HEADER *data, unsigned msgSize, unsigned char *message, unsigned *alertID);

// Function Declarats for Data Block Management
HRESULT componentCheck(ALERT_TRACKER *a_tracker);
HRESULT insertBlob(ALERT_TRACKER *a_tracker);
HRESULT insertComponentBlock (ALERT_TRACKER *a_tracker);
HRESULT linkNormToData (unsigned norm_id, unsigned data_id);
HRESULT renderVerdict(JUDGEMENT *verdict);
HRESULT getDataByAlertID(unsigned type, unsigned char **data, unsigned *size, int id);

#endif /*RZB_ALERT_DB_H*/

