#ifndef NRT_GLOBAL_H
#define NRT_GLOBAL_H

#include <stdio.h>
#include <netinet/in.h>

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <uuid/uuid.h>

#define URLLEN 1000
#define HOSTNAMELEN 256

#define NO_SUCCESSFUL_DETECTION 256
#define EXPLOIT                 4
#define SUSPICIOUS              3
#define WARNING                 2
#define ADVISORY                1
#define NO_ALERT                0

#define NUMSTREAMSTOTRACK 5000

#define NRTSID                  0xa5a5a5a5
#define INVALIDSTREAMIDX        0xFFFFFFFF
#define OUTOFSTREAMINFOSTORAGE  0xFFFFFFFF
#define INTEGER_OVERFLOW        0x41414141

#define SIZE_OF_UUID_STRING  37
#define MAX_DESCRIPTION_SIZE 1048

#ifndef ULONG_MAX
    #define ULONG_MAX   0xFFFFFFFF //4294967295UL // ULONG_MAX from limits.h
#endif

/* Well known Nugget Information */

/* Registration request size */
#define MAX_REG_SIZE 350
/* Nugget type: Correlation */
UUID_DEFINE(CORRELATION, 0x2f, 0xd7, 0x5f, 0xa5, 0x77, 0x8b, 0x44, 0x3e, 0xb9, 0x10, 0x1e, 0x19, 0x04, 0x4e, 0x81, 0xe1);
/* Nugget type: Detection */
UUID_DEFINE(DETECTION, 0xd9, 0x5a, 0xee, 0x72, 0x91, 0x86, 0x42, 0x36, 0xbf, 0x23, 0x8f, 0xf7, 0x7d, 0xac, 0x63, 0x0b);
/* Nugget type: Output */
UUID_DEFINE(OUTPUT, 0xa3, 0xd0, 0xd1, 0xf9, 0xc0, 0x49, 0x47, 0x4e, 0xbf, 0x01, 0x21, 0x28, 0xea, 0x00, 0xa7, 0x51);
/* Nugget type: Data Collector */
UUID_DEFINE(COLLECTOR, 0xc3, 0x8b, 0x11, 0x3a, 0x27, 0xfd, 0x41, 0x7c, 0xb9, 0xfa, 0xf3, 0xaa, 0x0a, 0xf5, 0xcb, 0x53);
/* Nugget type: Intel */
UUID_DEFINE(INTEL, 0x35, 0x61, 0x12, 0xd8, 0xf4, 0xf1, 0x41, 0xdc, 0xb3, 0xf7, 0xca, 0xce, 0x56, 0x74, 0xc2, 0xec);

/* Remove these once nuggets have been updated */
#define TESTCOLLECT  (const unsigned char *)"\xec\x4a\x83\xcc\x7e\x92\x4d\xbf\xa3\x64\xb9\x0f\xf9\xf6\x05\x87" /* SF - Test Collector v.1 */
#define TESTDETECT   (const unsigned char *)"\xc0\xc6\x3a\x39\x26\x3f\x44\x3e\x98\x5a\x7a\x9d\x1c\xa3\xc7\x27" /* SF - Test Detector v.1 */
#define TESTDETECT2  (const unsigned char *)"\x90\xad\x2e\xd4\x69\xba\x11\xdf\x84\x25\x33\x60\x9f\xdc\x13\x02" /* SF - TEST Detector v.2 */
#define NUG_CLAMAV   (const unsigned char *)"\xbe\x31\x20\x18\x1b\x8e\x46\x2e\x90\xa8\x98\x3a\x03\x6c\x8e\x8c"

/* Well know Data Types */
#define NO_DATA_TYPE (const unsigned char *)"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
#define PDF_FILE     (const unsigned char *)"\x00\x5d\x54\x64\x7a\x44\x49\x07\xaf\x57\x4d\xb0\x8a\x61\xe1\x3c"
#define SWF_FILE     (const unsigned char *)"\x7a\xb4\x5f\xff\x7c\x73\x41\x2c\x8b\x86\xc0\x76\x19\xc8\xfc\x7d"
#define ZIP_FILE     (const unsigned char *)"\x52\xf7\x38\xc2\xdb\x1c\x42\xca\xb5\xef\x50\xa4\xba\x3f\x75\x27"
#define PE_FILE      (const unsigned char *)"\xba\x9b\xeb\x5f\x06\x53\x4b\x04\x95\x52\x3b\xfb\x63\x4c\xa7\xfc"
#define OLE_DOC      (const unsigned char *)"\x16\xd7\x29\x48\x3d\x2b\x52\xed\x9a\xe4\x43\xef\x19\xce\x3e\x69"
#define MAIL_CAPTURE (const unsigned char *)"\xd1\x47\xf2\x15\x12\x8e\x47\x46\xa1\xe2\xb6\xc9\x78\xbb\x18\x69"
#define WEB_CAPTURE  (const unsigned char *)"\x5d\x1a\x26\xc7\xff\xa2\x40\x67\x9a\xf8\x4e\xd9\xee\x39\x38\x05"
#define DNS_CAPTURE  (const unsigned char *)"\x1d\x69\x35\x1c\xbc\x4c\x56\xe1\x81\x40\x0d\x83\x84\xe6\x14\x77"
#define ALERT_OUTPUT (const unsigned char *)"\x9b\xfc\x66\x6d\xc3\xd8\x55\xcc\xa2\xa5\x4d\x66\xd5\xa5\x0c\x59"
#define CAP_DESCRIP  (const unsigned char *)"\x3d\x39\x89\x3e\xae\x97\x53\x37\xac\x35\x0a\xbe\x30\x75\xfb\x88"
#define SHELLCODE    (const unsigned char *)"\x4e\x72\xc8\xec\xff\x88\x43\x71\xa0\xf0\xdf\xe2\xb4\xc7\x33\xdc"

/* Known nugget types */
#define SENDMAIL_NUGGET       (const unsigned char *)"\x1a\x11\x59\xba\xaf\xb2\x11\xdf\x8a\x74\xd3\x72\x38\xe2\x61\xd7" // Sendmail Milter Integrated Collector
#define SWF_NUGGET            (const unsigned char *)"\x20\xf7\x53\x80\xaf\xb0\x11\xdf\x86\x6f\x7b\x86\xc6\x49\x59\x6b" // Flash Inspector
#define LIBEMU_NUGGET         (const unsigned char *)"\x40\x56\x60\x82\xaf\xaf\x11\xdf\x9e\xd5\xe3\xab\x45\xe0\x14\xa4" // LibEMU Shellcode Detector
#define OFFICECAT_NUGGET      (const unsigned char *)"\x73\x39\x2f\x48\xaf\xaf\x11\xdf\xa5\x19\x8b\x58\xa5\x42\xa7\xe9" // OfficeCat
#define SMTP_PARSER_NUGGET    (const unsigned char *)"\x7a\xb0\x66\x9b\x37\x82\x58\x88\xb4\x51\x34\x79\xbf\xb6\xcd\xd4" // SMTP Parser
#define PDF_DISSECTOR_NUGGET  (const unsigned char *)"\x8d\x41\x2e\x36\xaf\xaf\x11\xdf\xb1\x23\xfb\xc1\x42\xfa\x74\x9e" // Zynamics PDF Dissector
#define TEST_DETECTOR2_NUGGET (const unsigned char *)"\x90\xad\x2e\xd4\x69\xba\x11\xdf\x84\x25\x33\x60\x9f\xdc\x13\x02" // TEST Detector v.2
#define VIRUSTOTAL_NUGGET     (const unsigned char *)"\x92\x6b\x5b\xea\xaf\xb2\x11\xdf\x80\x07\xfb\x6c\xd1\x23\x6c\x62" // Virus Total Integrated Inspector
#define SQUID_NUGGET          (const unsigned char *)"\x9d\x98\xbf\x86\xaf\xb1\x11\xdf\xbb\x5e\x2f\x25\x9b\x27\xf5\xea" // Squid Integrated Collector
#define SAAC_NUGGET           (const unsigned char *)"\xa3\xb1\x24\xf9\xee\x08\x5f\xdf\x8f\x3e\x5e\x8d\xdd\xe0\x83\xd5" // Snort As A Collector (SAAC)
#define FSMONITOR_NUGGET      (const unsigned char *)"\xb1\x85\x6e\x90\xaf\xb1\x11\xdf\xa7\xb9\xdb\x6c\xe9\x2e\x17\xa0" // File System Monitor Collector
#define CLAMAV_NUGGET         (const unsigned char *)"\xbe\x31\x20\x18\x1b\x8e\x46\x2e\x90\xa8\x98\x3a\x03\x6c\x8e\x8c" // Integrated ClamAV 0.96.1
#define TEST_DETECTOR_NUGGET  (const unsigned char *)"\xc0\xc6\x3a\x39\x26\x3f\x44\x3e\x98\x5a\x7a\x9d\x1c\xa3\xc7\x27" // Test Detector v.1
#define STEGDETECT_NUGGET     (const unsigned char *)"\xe3\xbe\x7a\x1e\xaf\xb1\x11\xdf\x8c\x80\x8b\x75\xaf\x31\x5b\x73" // Stegdetect Detection Nugget
#define TEST_COLLECTOR_NUGGET (const unsigned char *)"\xec\x4a\x83\xcc\x7e\x92\x4d\xbf\xa3\x64\xb9\x0f\xf9\xf6\x05\x87" // Test Collector v.1
#define POSTFIX_NUGGET        (const unsigned char *)"\xfe\xa5\x1a\x18\xaf\xb1\x11\xdf\xba\x4b\x9f\xff\xfd\x3f\xaf\xdb" // PostFix Integrated Collector
#define FSWALK_NUGGET         (const unsigned char *)"\xfb\xd7\x3e\x9a\xb5\x21\x11\xdf\x98\x21\x00\x1c\x23\x09\xdb\xa9" // File System Walker Collector

/* Standard Return Codes */
typedef enum
{
    R_NOT_FOUND = 333,
    R_FOUND = 444,
    R_SUCCESS = 555,
    R_FAIL = -1,
    R_MALLOC_FAIL = -2,
    R_ADD_FAIL = -3,
    R_DB_FAIL = -4,
    R_DUP_UID = -5,
    R_NO_ROUTE_TABLE = 4141,
    R_BUSY = -53,
} HRESULT;

/* Cache Codes */
#define CACHE_HIT_KNOWN_BAD   1
#define CACHE_HIT_KNOWN_GOOD  2

/* Network Cache Responsees */
typedef enum
{
    KNOWN_GOOD_RESPONSE = 2,
    KNOWN_BAD_RESPONSE = 1,
    UNKNOWN_MD5_RESPONSE = 0,
} NetworkCacheResponse;

/* Network Message Size Defines */
#define MAX_IP_SIZE           16   /* vvv.www.xxx.yyy + \x00 */
#define MAX_PORT_SIZE         6    /* 65535 + \x00 */

#define SAFE_MD5_SIZE   33
#define MD5_SIZE        16

typedef enum
{
    REQ, ACK, RDR, REG, POLL, RST, MD5CHECK, MD5RESPONSE, META, REFERRAL,
    ALRT, DETECTDONE, NEWCAP, ERR, REQBLOB, SENDBLOB, NTEL, WKST, VERDICT
} MESSAGE;

typedef enum
{
    CAP_DETAILS
} METATYPE;

/*
 *
 * NRT CONFIG
 *
 */
typedef enum
{
    GLOBAL, DSRV, NSRV, DBASE, OUTPUTNUG
} CONFSECTION;

typedef struct _NRTConfig
{
//Tracking
    void *currval;
    CONFSECTION section;
//Global
    int maximumthreads;
    int badcachesize;
    int goodcachesize;
    int urlcachesize;
//Nugget Server
    char *nugname;
    char *nugport;
    char *nugaddr;
    char *handlerdir;
//Dispatch Server
    int nuggettype;
    int nummachines;
    char *dsrvaddr;
    char *dsrvport;
    int routingtype;
    int noalert_flags;
    int drop_flags;
//Databse info
    char *db;
    char *dbhost;
    char *username;
    char *password;
// Output nuggets
    char *on1_ip;
    char *on1_port;
    char *on2_ip;
    char *on2_port;
} NRTConfig;


/*
 *
 * PROTOCOL
 *
 */

typedef struct _SNAKECHARMER
{
    char magic[3];
    uint8_t version;
    MESSAGE msgtype;
    uuid_t datatype;
} SNAKECHARMER;

typedef struct _REGPACKET
{
    unsigned nuggetid;
    struct in_addr nugaddr;
    unsigned short nugport;
    uuid_t nugtype;
    uuid_t nugapp;
    uuid_t datatype;
    unsigned short dummy;
    unsigned freeinstances;
    char name[256];
} REGPACKET;

typedef struct _NEWCAPPACKET
{
    unsigned nuggetid;
    struct in_addr nugaddr;
    unsigned short nugport;
    uuid_t nugtype;
    uuid_t nugapp;
    uuid_t datatype;
} NEWCAPPACKET;

// The parent_md5 is used to track sub-components in the database.
// For example, if we receive a PDF file md5:a, and it has a JavaScript
// block md5:b and that block has a shellcode block md5:c, the transit would be
// as follows:
// reqpacket: md5:a, p_md5:a <-- Matching md5/p_md5 means this is a top-level block
// reqpacket: md5:b, p_md5:a <-- The javascript is a child to the PDF
// regpacket: md5:c, p_md5:b <-- The shellcode is a child to the JavaScript.
// In this way, from the database we can describe relationships between components:
//
// a <- b <- c
//
// If we had another PDF file md5:d, with the same javascript block, we could then
// describe the relationships as:
//
// a <-|
//     |--b <- c
// d <-|
//
// This is useful from a forensics and Intelligence Driven Response perspective
typedef struct _REQPACKET
{
    unsigned      eventid;             // Temporary id, resoved by dispatcher on alert
    unsigned      timestamp;           // Seconds since epoch
    unsigned char     md5[MD5_SIZE];             // MD5 of current block
    unsigned      filesize;
    uuid_t            datatype;
    unsigned char     parent_md5[MD5_SIZE];      // MD5 of parent block
    unsigned      parent_size;
    char              filename[URLLEN];    // Optional
    unsigned      src_ip;
    unsigned      dst_ip;
    unsigned short    ip_proto;            // Optional
    unsigned short    src_port;            // Optional
    unsigned short    dst_port;            // Optional
} REQPACKET;

typedef struct _BLOCK_META_DATA
{
    unsigned eventid;
    unsigned timestamp;           // Seconds since epoch
    uint8_t *data;
    unsigned size;
    uuid_t datatype;
    unsigned char parent_md5[MD5_SIZE];      // If this is 00
    void *parent_data;        // "NULL" means this is top-level datablock
    unsigned parent_size;         // Ignored if parent_data is "NULL"
    struct in_addr src_ip;
    struct in_addr dst_ip;
    unsigned short ip_proto;            // Optional
    unsigned short src_port;            // Optional
    unsigned short dst_port;            // Optional
} BLOCK_META_DATA;

typedef struct _MD5CACHEREQUEST
{
    unsigned filesize;
    unsigned char md5[MD5_SIZE];
    uuid_t datatype;
} MD5CACHEREQUEST;

typedef struct _MD5CACHERESPONSE
{
    uint32_t result;
    uint32_t detectionID;
    uint32_t referralcount;
} MD5CACHERESPONSE;

typedef struct _DETECTIONREFERRAL
{
    char ip[MAX_IP_SIZE];
    char port[MAX_PORT_SIZE];
} DETECTIONREFERRAL;

typedef struct _ALERTPACKET
{
    unsigned alertid;
    unsigned priority;
    unsigned alertsize;
    //short alert
} ALERTPACKET;

typedef struct _METAPACKET
{
    unsigned type;
    unsigned filesize;
    unsigned char md5[MD5_SIZE];
    unsigned size;
} METAPACKET;


/*
 *
 * NETWORK INFO
 *
 */
struct _RZBNetworkConnection;
typedef struct _RZBNetworkConnection RZBNetworkConnection;

/*
 *
 * THREADING INFO
 *
 */
typedef struct _THREADARGS
{
    RZBNetworkConnection *net;
    unsigned threadindex;
    unsigned nuggetid;
    unsigned eventid;
    unsigned char *data;
    unsigned size;
    uuid_t type;
    BLOCK_META_DATA *metaData;
} THREADARGS;

typedef void *(*rzb_thread_func_t)(THREADARGS *);

/*
 *
 * INTERFACE OBJECTS
 *
 */
typedef void (*handler_fp)(BLOCK_META_DATA *metData);
typedef HRESULT (*reg_handler_fp)(handler_fp fp,
                                  const uuid_t *acceptedTypes,
                                  size_t numTypes,
                                  const uuid_t nuggetType );
typedef HRESULT (*send_alert_fp)( unsigned alertid,
                                  char *alertmsg,
                                  unsigned metacode,
                                  unsigned metasize,
                                  void  *metadata );

typedef HRESULT (*send_blockdata_fp)( BLOCK_META_DATA *metaData );

typedef HRESULT (*alerts_done_fp)( unsigned eventid );

typedef HRESULT (*send_metadata_fp)( uuid_t *id,
                                     unsigned char *mdData,
                                     unsigned size,
                                     uuid_t *type );

typedef struct _DETECTIONAPI
{
    reg_handler_fp      registerHandler;
    send_alert_fp       sendAlert;
    send_blockdata_fp   sendData;
    alerts_done_fp      AlertsDone;
    send_metadata_fp    sendMetaData;
} DetectionAPI;

typedef HRESULT (*initNugFunc)(const DetectionAPI *);

typedef struct _COLLECTORINFO
{
    uuid_t nuggetType;     // nugget application type - collector is assumed
    char name[256];
} CollectorInfo;

typedef HRESULT (*check_resource_fp)(const uuid_t id, const char *data, const uuid_t type);
typedef HRESULT (*send_data_fp)(BLOCK_META_DATA *metaData);
typedef HRESULT (*send_meta_data_fp)(METAPACKET *mpacket, unsigned char *data);
typedef HRESULT (*register_nugget_fp)(const uuid_t   nugtype,
                                      const uuid_t   nugapp,
                                      const uuid_t   datatype,
                                      unsigned short nugport,
                                      unsigned   instances,
                                      unsigned   nugid,
                                      const char     *name,
                                      unsigned   *nuggetid);

#endif /* NRT_GLOBAL_H */
